from importlib.metadata import version as get_version

__version__ = get_version(__package__)

__version_info__ = tuple(
    int(num) if num.isdigit() else num
    for num in __version__.replace("-", ".", 1).split(".")
)



"""_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_myproj gRPC handlers*

:details: lara_django_myproj gRPC handlers.
         - 
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: - 
________________________________________________________________________
"""

# generated with django-socio-grpc generateprpcinterface lara_django_organisms  (LARA-version)

# import logging
from django_socio_grpc.services.app_handler_registry import AppHandlerRegistry
from lara_django_organisms.grpc.services import (
    ExtraDataService,
    HabitatClassService,
    HabitatParameterService,
    HabitatService,
    TraitService,
    DomainService,
    RegnumService,
    PhylumService,
    ClassisService,
    OrdoService,
    FamiliaService,
    GenusService,
    SpeciesService,
    SubspeciesService,
    VariantService,
    CultivarService,
    OrganismService,
    OrganismsSubsetService
)


def grpc_handlers(server):
    app_registry = AppHandlerRegistry("lara_django_organisms", server)

    app_registry.get_grpc_module = lambda: "lara_django_organisms_grpc.v1"

    app_registry.register(ExtraDataService)

    app_registry.register(HabitatClassService)

    app_registry.register(HabitatParameterService)

    app_registry.register(HabitatService)

    app_registry.register(TraitService)

    app_registry.register(DomainService)

    app_registry.register(RegnumService)

    app_registry.register(PhylumService)

    app_registry.register(ClassisService)

    app_registry.register(OrdoService)

    app_registry.register(FamiliaService)

    app_registry.register(GenusService)

    app_registry.register(SpeciesService)

    app_registry.register(SubspeciesService)

    app_registry.register(VariantService)

    app_registry.register(CultivarService)

    app_registry.register(OrganismService)

    app_registry.register(OrganismsSubsetService)



"""_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_organisms admin *

:details: lara_django_organisms admin module admin backend configuration.
         - 
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: - run "lara-django-dev admin_generator lara_django_organisms >> admin.py" to update this file
________________________________________________________________________
"""
# -*- coding: utf-8 -*-
from django.contrib import admin

from .models import (
    ExtraData,
    HabitatClass,
    HabitatParameter,
    Habitat,
    Trait,
    Domain,
    Regnum,
    Phylum,
    Classis,
    Ordo,
    Familia,
    Genus,
    Species,
    Subspecies,
    Variant,
    Cultivar,
    Organism,
    OrganismsSubset,
)


@admin.register(ExtraData)
class ExtraDataAdmin(admin.ModelAdmin):
    list_display = (
        "namespace",
        "data_type",
        "media_type",
        "iri",
        "url",
        "description",
        "file",
        "extradata_id",
    )
    list_filter = ("data_type", "namespace", "media_type")


@admin.register(HabitatClass)
class HabitatClassAdmin(admin.ModelAdmin):
    list_display = ("habitatclass_id", "name", "description")
    search_fields = ("name",)


@admin.register(HabitatParameter)
class HabitatParameterAdmin(admin.ModelAdmin):
    list_display = (
        "habitatparameter_id",
        "name",
        "unit",
        "description",
    )
    list_filter = ("unit",)
    raw_id_fields = ("data_extra",)
    search_fields = ("name",)


@admin.register(Habitat)
class HabitatAdmin(admin.ModelAdmin):
    list_display = ("habitat_id", "name", "description")
    raw_id_fields = ("habitat_class", "parameters")
    search_fields = ("name",)


@admin.register(Trait)
class TraitAdmin(admin.ModelAdmin):
    list_display = ("trait_id", "name", "unit", "description")
    list_filter = ("unit",)


@admin.register(Domain)
class DomainAdmin(admin.ModelAdmin):
    list_display = ("domain_id", "name", "description")
    search_fields = ("name",)


@admin.register(Regnum)
class RegnumAdmin(admin.ModelAdmin):
    list_display = ("regnum_id", "name", "description")
    search_fields = ("name",)


@admin.register(Phylum)
class PhylumAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "name_common",
        "characteristics",
        "description",
        "phylum_id",
    )
    search_fields = ("name",)


@admin.register(Classis)
class ClassisAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "name_common",
        "characteristics",
        "description",
        "classis_id",
    )
    search_fields = ("name",)


@admin.register(Ordo)
class OrdoAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "name_common",
        "characteristics",
        "description",
        "ordo_id",
    )
    search_fields = ("name",)


@admin.register(Familia)
class FamiliaAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "name_common",
        "characteristics",
        "description",
        
    )
    search_fields = ("name",)


@admin.register(Genus)
class GenusAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "name_common",
        "characteristics",
        "description",
        "genus_id",
    )
    search_fields = ("name",)


@admin.register(Species)
class SpeciesAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "name_common",
        "characteristics",
        "description",
        "species_id",
    )
    search_fields = ("name",)


@admin.register(Subspecies)
class SubspeciesAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "name_common",
        "characteristics",
        "description",
        "subspecies_id",
    )
    search_fields = ("name",)


@admin.register(Variant)
class VariantAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "name_common",
        "characteristics",
        "description",
        "variant_id",
    )
    search_fields = ("name",)


@admin.register(Cultivar)
class CultivarAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "name_common",
        "characteristics",
        "description",
        "cultivar_id",
    )
    search_fields = ("name",)


@admin.register(Organism)
class OrganismAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "name_common",
        "domain",
        "regnum",
        "phylum",
        "classis",
        "ordo",
        "familia",
        "genus",
        "species",
        "subspecies",
        "variant",
        "cultivar",
        "description",
        "organism_id",
    )
    list_filter = (
        "domain",
        "regnum",
        "phylum",
        "classis",
        "ordo",
        "familia",
        "genus",
        "species",
        "subspecies",
        "variant",
        "cultivar",
        "habitats",
    )
    raw_id_fields = ("traits", "data_extra")
    search_fields = ("name",)

@admin.register(OrganismsSubset)
class OrganismsSubsetAdmin(admin.ModelAdmin):
    list_display = (
        "namespace",
        "name",
        "description",
        "organismssubset_id",
    )
    list_filter = ("namespace",)
    raw_id_fields = ("organisms",)
    search_fields = ("name",)


"""_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_organisms admin *

:details: lara_django_organisms admin module admin backend configuration.
         - 
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: - run "lara-django-dev tables_generator lara_django_organisms >> tables.py" to update this file
________________________________________________________________________
"""
# django Tests s. https://docs.djangoproject.com/en/4.1/topics/testing/overview/ for lara_django_organisms
# generated with django-extensions tests_generator  lara_django_organisms > tests.py (LARA-version)

import logging
from django.utils.html import format_html
import django_tables2 as tables


from .models import (
    ExtraData,
    HabitatClass,
    HabitatParameter,
    Habitat,
    Trait,
    Domain,
    Regnum,
    Phylum,
    Classis,
    Ordo,
    Familia,
    Genus,
    Species,
    Subspecies,
    Variant,
    Cultivar,
    Organism,
)


class ExtraDataTable(tables.Table):
    # adding link to column <column-to-be-linked>
    # name = tables.Column(linkify=('lara_django_organisms:extradata-detail', [tables.A('pk')]))

    class Meta:
        model = ExtraData

        fields = (
            "data_type",
            "namespace",
            "uri",
            "media_type",
            "iri",
            "url",
            "description",
            "file",
        )


class HabitatClassTable(tables.Table):
    # adding link to column <column-to-be-linked>
    name = tables.Column(
        linkify=("lara_django_organisms:habitatclass-detail", [tables.A("pk")])
    )

    class Meta:
        model = HabitatClass

        fields = ("name", "description")


class HabitatParameterTable(tables.Table):
    # adding link to column <column-to-be-linked>
    name = tables.Column(
        linkify=("lara_django_organisms:habitatparameterclass-detail", [tables.A("pk")])
    )

    class Meta:
        model = HabitatParameter

        fields = ("name", "unit", "description")


class HabitatTable(tables.Table):
    # adding link to column <column-to-be-linked>
    name = tables.Column(
        linkify=("lara_django_organisms:habitat-detail", [tables.A("pk")])
    )

    class Meta:
        model = Habitat

        fields = ("name", "description")


class TraitTable(tables.Table):
    # adding link to column <column-to-be-linked>
    name = tables.Column(
        linkify=("lara_django_organisms:featureclass-detail", [tables.A("pk")])
    )

    class Meta:
        model = Trait

        fields = ("feature", "unit", "description")


class DomainTable(tables.Table):
    # adding link to column <column-to-be-linked>
    name = tables.Column(
        linkify=("lara_django_organisms:domain-detail", [tables.A("pk")])
    )

    class Meta:
        model = Domain

        fields = ("name", "description")


class RegnumTable(tables.Table):
    # adding link to column <column-to-be-linked>
    name = tables.Column(
        linkify=("lara_django_organisms:regnum-detail", [tables.A("pk")])
    )

    class Meta:
        model = Regnum

        fields = ("name", "description")


class PhylumTable(tables.Table):
    # adding link to column <column-to-be-linked>
    name = tables.Column(
        linkify=("lara_django_organisms:phylum-detail", [tables.A("pk")])
    )

    class Meta:
        model = Phylum

        fields = ("name", "name_common", "characteristics", "description")


class ClassisTable(tables.Table):
    # adding link to column <column-to-be-linked>
    name = tables.Column(
        linkify=("lara_django_organisms:classis-detail", [tables.A("pk")])
    )

    class Meta:
        model = Classis

        fields = ("name", "name_common", "characteristics", "description")


class OrdoTable(tables.Table):
    # adding link to column <column-to-be-linked>
    name = tables.Column(
        linkify=("lara_django_organisms:ordo-detail", [tables.A("pk")])
    )

    class Meta:
        model = Ordo

        fields = ("name", "name_common", "characteristics", "description")


class FamiliaTable(tables.Table):
    # adding link to column <column-to-be-linked>
    name = tables.Column(
        linkify=("lara_django_organisms:familia-detail", [tables.A("pk")])
    )

    class Meta:
        model = Familia

        fields = ("name", "name_common", "characteristics", "description")


class GenusTable(tables.Table):
    # adding link to column <column-to-be-linked>
    name = tables.Column(
        linkify=("lara_django_organisms:genus-detail", [tables.A("pk")])
    )

    class Meta:
        model = Genus

        fields = ("name", "name_common", "characteristics", "description")


class SpeciesTable(tables.Table):
    # adding link to column <column-to-be-linked>
    name = tables.Column(
        linkify=("lara_django_organisms:species-detail", [tables.A("pk")])
    )

    class Meta:
        model = Species

        fields = ("name", "name_common", "characteristics", "description")


class SubspeciesTable(tables.Table):
    # adding link to column <column-to-be-linked>
    name = tables.Column(
        linkify=("lara_django_organisms:subspecies-detail", [tables.A("pk")])
    )

    class Meta:
        model = Subspecies

        fields = ("name", "name_common", "characteristics", "description")


class VariantTable(tables.Table):
    # adding link to column <column-to-be-linked>
    name = tables.Column(
        linkify=("lara_django_organisms:variant-detail", [tables.A("pk")])
    )

    class Meta:
        model = Variant

        fields = ("name", "name_common", "characteristics", "description")


class CultivarTable(tables.Table):
    # adding link to column <column-to-be-linked>
    name = tables.Column(
        linkify=("lara_django_organisms:cultivar-detail", [tables.A("pk")])
    )

    class Meta:
        model = Cultivar

        fields = ("name", "name_common", "characteristics", "description")


class OrganismTable(tables.Table):
    # adding link to column <column-to-be-linked>
    name = tables.Column(
        linkify=("lara_django_organisms:organism-detail", [tables.A("pk")])
    )

    edit = tables.Column(
        empty_values=(),
        verbose_name="Edit",
        default="Edit",
        linkify={
            "viewname": "lara_django_organisms:organism-update",
            "args": [(tables.A("pk"))],
        },
        attrs={"a": {"class": "btn btn-warning btn-sm"}},
    )

    delete = tables.Column(
        empty_values=(),
        verbose_name="Delete",
        default="Delete",
        linkify={
            "viewname": "lara_django_organisms:organism-delete",
            "args": [(tables.A("pk"))],
        },
        attrs={"a": {"class": "btn btn-danger btn-sm"}},
    )

    def render_edit(self):
        return format_html('<i class="bi bi-pencil-square"></i>')

    def render_delete(self):
        return format_html('<i class="bi bi-trash"></i>')

    class Meta:
        model = Organism

        fields = (
            "name",
            "name_common",
            "domain",
            "regnum",
            "phylum",
            "classis",
            "ordo",
            "familia",
            "genus",
            "species",
            "subspecies",
            "variant",
            "cultivar",
            "habitat",
            "edit",
            "delete",
        )



"""_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_organisms models *

:details: lara_django_organisms database models.
         -
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: -
________________________________________________________________________
"""

import uuid
import hashlib
import json
import logging
from django.utils import timezone
from django.conf import settings
from django.utils.translation import gettext_lazy as _

from django.db import models


from lara_django_base.models import ExtraDataAbstr, PhysicalUnit, Tag

from lara_django_people.models import Entity, LaraUser, ItemSubsetAbstr


settings.FIXTURES += [
    "1_habitat_class_fix",
    "2_domain_fix",
    "3_regnum_fix",
    # "4_phylum_fix",
    # "5_classis_fix",
    # "6_ordo_fix",
    "7_familia_fix",
    "8_genus_fix",
    # "9_species_fix",
    # "10_subspecies_fix",
    # "11_variant_fix",
    # "12_cultivar_fix",
]

logger = logging.getLogger(__name__)


class ExtraData(ExtraDataAbstr):
    """This class can be used to extend data, by extra information,
    e.g. more telephone numbers, customer numbers, ..."""

    model_curi = "lara:organisms/ExtraData"
    file = models.FileField(
        upload_to="organisms", blank=True, null=True, help_text="rel. path/filename"
    )

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for name_full
        """

        if self.version is None or self.version == "":
            self.version = "0.0.1"

        if self.hash_sha256 is None:
            if self.data_json is not None:
                to_hash = json.dumps(self.data_json)  # +  str(self.UUID) +
                self.hash_sha256 = hashlib.sha256(to_hash.encode("utf-8")).hexdigest()
            else:
                self.hash_sha256 = hashlib.sha256(
                    str(self.extradata_id).encode("utf-8")
                ).hexdigest()

        if self.name is None or self.name == "":
            if self.name_display is None or self.name_display == "":
                self.name = "_".join(
                    (
                        timezone.now().strftime("%Y%m%d_%H%M%S"),
                        "data",
                        self.hash_sha256[:8],
                    )
                )
            else:
                self.name = self.name_display.replace(" ", "_").lower().strip()

        else:
            self.name = self.name.replace(" ", "_").lower().strip()
        if self.name_full is None or self.name_full == "":
            if self.namespace is not None and self.version is not None:
                self.name_full = f"{self.namespace.uri}/" + "_".join(
                    (self.name.replace(" ", "_"), self.version)
                )
            else:
                self.name_full = "_".join((self.name.replace(" ", "_"), self.version))

        if not self.iri:
            netloc = self.namespace.parse_URI().netloc
            path = self.namespace.parse_URI().path

            self.iri = f"{settings.LARA_PREFIXES['lara']}organisms/ExtraData/{netloc}{path}/{self.name.replace(' ', '_').lower().strip()}"

        self.datetime_last_modified = timezone.now()

        super().save(*args, **kwargs)


class HabitatClass(models.Model):
    """classes of habitats, like maritime, terrestrial, shore, .."""

    model_curi = "lara:organisms/HabitatClass"
    habitatclass_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False
    )
    name = models.TextField(unique=True, help_text="maritime, terrestrial, shore")
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    description = models.TextField(
        blank=True, help_text="description of the habitat class"
    )

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for IRI
        """
        if self.iri is None or self.iri == "":
            self.iri = f"{settings.LARA_PREFIXES['lara']}organisms/HabitatClass/{self.name.replace(' ', '_').lower().strip()}"

        super().save(*args, **kwargs)

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return self.name or ""

    class Meta:
        verbose_name_plural = "HabitatClasses"
        # ~ db_table = "lara_substances_substance_class"


class HabitatParameter(models.Model):
    """Class level Habitat Parameters can be used to model all kinds of general habitat parameters
    that are common for all organisms of a certain class.
    Habitat parameters of an individual organism should be captured in the
    organism store.
    Examples of these parameters could be moisture, pH, nutrients (?), geological,
     geo_width_min, geo_width_max
     temperature_average
     temperature_min
     temperature_max
     oxygen_concentration_min
     oxygen_concentration_max
     oxygen_concentration_average
     humidity_average
     pH_average

     # salt concentration
     moisture_air_min
     moisture_air_max
     moisture_air_average
     light_during_day_min
     light_during_day_max
     light_during_day_average

     height_min
     height_max
    :TODO: parts should be moved to ontology
    """

    model_curi = "lara:organisms/HabitatParameter"
    habitatparameter_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False
    )
    name_display = models.TextField(
        help_text="human readable display name of the habitat parameter, can contain spaces"
    )
    name = models.TextField(
        unique=True,
        help_text="name of the habitat parameter, e.g. temperature, moisture, pH, ...",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    value_range = models.JSONField(
        null=True,
        blank=True,
        help_text="range of values for this parameter, like min, max, average, ...",
    )
    unit = models.ForeignKey(
        PhysicalUnit,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_unit_related",
        related_query_name="%(app_label)s_%(class)s_unit_related_query",
        blank=True,
        null=True,
        help_text="physical unit of the parameter, e.g. K, mol/L, ...",
    )
    data_extra = models.ManyToManyField(
        ExtraData,
        related_name="%(app_label)s_%(class)s_data_extra_related",
        related_query_name="%(app_label)s_%(class)s_data_extra_related_query",
        blank=True,
        help_text="e.g. more information about this particular parameter ...",
    )
    description = models.TextField(
        null=True, blank=True, help_text="description of the habitat parameter"
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for IRI
        """
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()

        if self.iri is None or self.iri == "":
            self.iri = f"{settings.LARA_PREFIXES['lara']}organisms/HabitatParameter/{self.name.replace(' ', '_').lower().strip()}"

        self.datetime_last_modified = timezone.now()

        super().save(*args, **kwargs)

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return self.name or ""

    # class Meta:
    #     verbose_name_plural = "HabitatParameters"
    # ~ db_table = "lara_substances_substance_class"


class Habitat(models.Model):
    """Information about the living conditions of an organism, e.g. temperature, moisture, pH, ..."""

    model_curi = "lara:organisms/Habitat"
    habitat_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the habitat, can contain spaces"
    )
    name = models.TextField(
        blank=True,
        null=True,
        help_text="habitat name with namespaces, like sea.shore, sea.deepsee, land.vulcan",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    habitat_class = models.ForeignKey(
        HabitatClass,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_habitat_class_related",
        related_query_name="%(app_label)s_%(class)s_habitat_class_related_query",
        blank=True,
        null=True,
        help_text="class of the habitat, like maritime, terrestrial, shore, ...",
    )
    parameters = models.ManyToManyField(
        HabitatParameter,
        blank=True,
        related_name="%(app_label)s_%(class)s_parameters_related",
        related_query_name="%(app_label)s_%(class)s_parameters_related_query",
        help_text="e.g. more parameters, like nutrients, ...",
    )
    description = models.TextField(
        null=True, blank=True, help_text="description of the habitat"
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat data was last modified",
    )

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for IRI
        """
        if self.iri is None or self.iri == "":
            self.iri = f"{settings.LARA_PREFIXES['lara']}organisms/habitat/{self.name.replace(' ', '_').lower().strip()}"

        self.datetime_last_modified = timezone.now()

        super().save(*args, **kwargs)

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return self.name or ""


class Trait(models.Model):
    """Phenotypical, phenomenological or genetic traits / features / attributes of the general organism,
    like branched, round, oval, .. (GRAM) staining, genetic fingerprint sequence,
    or genetic features, like antibiotic resistance,
     morphology
     size
     shape
     motility
     spore
     flagella
     capsule

     growth optimum temp, pH, salt, sugar, ...
     doubling rate
     metabolites
     nutrients
     metabolic characteristics
     pathogenicity
     membrane potential

     legal status / origin / patent
     part of the organism
     culture conditions
    """

    model_curi = "lara:organisms/Trait"
    trait_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the trait, can contain spaces"
    )
    name = models.TextField(
        unique=True,
        help_text="Name of the trait / feature attribute of the organisms, like branched, round, oval, .. staining, genetic fingerprint sequence ",
    )
    value_range = models.JSONField(
        null=True,
        blank=True,
        help_text="range of values for this trait, like min, max, average, ...",
    )
    characteristics = models.JSONField(
        blank=True, null=True, help_text="characteristics of this trait"
    )
    unit = models.ForeignKey(
        PhysicalUnit,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_unit_related",
        related_query_name="%(app_label)s_%(class)s_unit_related_query",
        blank=True,
        null=True,
        help_text="physical unit of the parameter, e.g. m, m^2, V, ...",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    description = models.TextField(
        null=True, blank=True, help_text="description of the feature class"
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when trait data was last modified",
    )

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for IRI
        """

        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()

        if self.iri is None or self.iri == "":
            self.iri = f"{settings.LARA_PREFIXES['lara']}organisms/FeatureClass/{self.name.replace(' ', '_').lower().strip()}"

        self.datetime_last_modified = timezone.now()
        self.datetime_last_modified = timezone.now()

        super().save(*args, **kwargs)

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return self.name or ""

    class Meta:
        verbose_name_plural = "FeatureClasses"


class Domain(models.Model):
    """taxonomic domain / super kingdom / empire: Archaea, Bacteria, and Eukarya"""

    model_curi = "lara:organisms/Domain"
    domain_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the domain, can contain spaces"
    )
    name = models.TextField(
        blank=True, null=True, help_text="domain name: Archaea, Bacteria, and Eukarya"
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_domain_traits_related",
        related_query_name="%(app_label)s_%(class)s_domain_traits",
        blank=True,
        help_text="General domain's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="General characteristics of Domain. Use 'traits' for specific traits.",
    )
    description = models.TextField(
        null=True, blank=True, help_text="description of domain"
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for IRI
        """

        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()

        if self.iri is None or self.iri == "":
            self.iri = f"{settings.LARA_PREFIXES['lara']}organisms/domain/{self.name.replace(' ', '_').lower().strip()}"
        self.datetime_last_modified = timezone.now()
        super().save(*args, **kwargs)

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return self.name or ""


class Regnum(models.Model):
    """taxonomic Regnum (Kingdom): Animalia, Plantae, Fungi, Protista,
    Archaea/Archaebacteria, and Bacteria/Eubacteria
    might be deprecated
    """

    model_curi = "lara:organisms/Regnum"
    regnum_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the regnum, can contain spaces"
    )
    name = models.TextField(
        blank=True, null=True, help_text="Regnum name, like Plantae, Fungi"
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_regnum_traits_related",
        related_query_name="%(app_label)s_%(class)s_regnum_traits",
        blank=True,
        help_text="General regnum's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="General characteristics of Regnum, use 'traits' for specific traits.",
    )
    description = models.TextField(
        null=True, blank=True, help_text="description of Regnum"
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for IRI
        """
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()

        if self.iri is None or self.iri == "":
            self.iri = f"{settings.LARA_PREFIXES['lara']}organisms/regnum/{self.name.replace(' ', '_').lower().strip()}"
        self.datetime_last_modified = timezone.now()
        super().save(*args, **kwargs)

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return self.name or ""

    class Meta:
        verbose_name_plural = "regna"


class Phylum(models.Model):
    """taxonomic Phylum / Devisio / Division"""

    model_curi = "lara:organisms/Phylum"
    phylum_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the phylum, can contain spaces"
    )
    name = models.TextField(blank=True, null=True, help_text="phylum name, like ")
    name_common = models.TextField(
        blank=True, null=True, help_text="common phylum name, like "
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_phylum_traits_related",
        related_query_name="%(app_label)s_%(class)s_phylum_traits",
        blank=True,
        help_text="General phylum's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="Generic characteristics of phylum. Use 'traits' for specific traits.",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    description = models.TextField(
        null=True, blank=True, help_text="description of phylum"
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for IRI
        """
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()
        if self.iri is None or self.iri == "":
            self.iri = f"{settings.LARA_PREFIXES['lara']}organisms/phylum/{self.name.replace(' ', '_').lower().strip()}"
        self.datetime_last_modified = timezone.now()
        super().save(*args, **kwargs)

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return self.name or ""

    class Meta:
        verbose_name_plural = "Phyla"


class Classis(models.Model):
    """taxonomic Classis / Class"""

    model_curi = "lara:organisms/Classis"
    classis_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the classis, can contain spaces"
    )
    name = models.TextField(blank=True, null=True, help_text="classis name, like ")
    name_common = models.TextField(
        blank=True, null=True, help_text="common classis name, like "
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_classis_traits_related",
        related_query_name="%(app_label)s_%(class)s_classis_traits",
        blank=True,
        help_text="General classis's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="generic characteristics of this classis, use 'traits' for specific traits.",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    description = models.TextField(
        null=True, blank=True, help_text="description of this classis"
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for IRI
        """
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()
        if self.iri is None or self.iri == "":
            self.iri = f"{settings.LARA_PREFIXES['lara']}organisms/classis/{self.name.replace(' ', '_').lower().strip()}"
        self.datetime_last_modified = timezone.now()
        super().save(*args, **kwargs)

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return self.name or ""

    class Meta:
        verbose_name_plural = "Classes"


class Ordo(models.Model):
    """taxonomic Ordo / order"""

    model_curi = "lara:organisms/Ordo"
    ordo_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the ordo, can contain spaces"
    )
    name = models.TextField(blank=True, null=True, help_text="ordo name, like ")
    name_common = models.TextField(
        blank=True, null=True, help_text="common ordo name, like "
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_ordo_traits_related",
        related_query_name="%(app_label)s_%(class)s_ordo_traits",
        blank=True,
        help_text="General ordo's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="Generic characteristics of this ordo, use 'traits' for specific traits.",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    description = models.TextField(
        null=True, blank=True, help_text="description of this ordo"
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for IRI
        """
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()
        if self.iri is None or self.iri == "":
            self.iri = f"{settings.LARA_PREFIXES['lara']}organisms/ordo/{self.name.replace(' ', '_').lower().strip()}"
        self.datetime_last_modified = timezone.now()
        super().save(*args, **kwargs)

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return self.name or ""

    class Meta:
        verbose_name_plural = "Ordae"


class Familia(models.Model):
    """taxonomic Familia / Family"""

    model_curi = "lara:organisms/Familia"
    familia_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the familia, can contain spaces"
    )
    name = models.TextField(blank=True, null=True, help_text="familia name, like ")
    name_common = models.TextField(
        blank=True, null=True, help_text="common familia name, like "
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_familia_traits_related",
        related_query_name="%(app_label)s_%(class)s_familia_traits",
        blank=True,
        help_text="General familia's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="Generic characteristics of this familia, use 'traits' for specific traits.",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    description = models.TextField(
        null=True, blank=True, help_text="description of this familia"
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for IRI
        """
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()
        if self.iri is None or self.iri == "":
            self.iri = f"{settings.LARA_PREFIXES['lara']}organisms/familia/{self.name.replace(' ', '_').lower().strip()}"
        self.datetime_last_modified = timezone.now()
        super().save(*args, **kwargs)

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return self.name or ""

    class Meta:
        verbose_name_plural = "Familiae"


class Genus(models.Model):
    """taxonomic Genus"""

    model_curi = "lara:organisms/Genus"
    genus_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the genus, can contain spaces"
    )
    name = models.TextField(blank=True, null=True, help_text="genus name, like ")
    name_common = models.TextField(
        blank=True, null=True, help_text="common genus name, like "
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_genus_traits_related",
        related_query_name="%(app_label)s_%(class)s_genus_traits",
        blank=True,
        help_text="General genus's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="General characteristics of this genus, use 'traits' for specific traits.",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    description = models.TextField(
        null=True, blank=True, help_text="description of this genus"
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for IRI
        """
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()
        if self.iri is None or self.iri == "":
            self.iri = f"{settings.LARA_PREFIXES['lara']}organisms/genus/{self.name.replace(' ', '_').lower().strip()}"
        self.datetime_last_modified = timezone.now()
        super().save(*args, **kwargs)

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return self.name or ""

    class Meta:
        verbose_name_plural = "Genera"


class Species(models.Model):
    """taxonomic Species"""

    model_curi = "lara:organisms/Species"
    species_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the species, can contain spaces"
    )
    name = models.TextField(blank=True, null=True, help_text="species name, like ")
    name_common = models.TextField(
        blank=True, null=True, help_text="common species name, like "
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_species_traits_related",
        related_query_name="%(app_label)s_%(class)s_species_traits",
        blank=True,
        help_text="General species's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="General characteristics of this species, use 'traits' for specific traits.",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    description = models.TextField(
        null=True, blank=True, help_text="description of this species"
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for IRI
        """
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()
        if self.iri is None or self.iri == "":
            self.iri = f"{settings.LARA_PREFIXES['lara']}organisms/species/{self.name.replace(' ', '_').lower().strip()}"
        self.datetime_last_modified = timezone.now()
        super().save(*args, **kwargs)

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return self.name or ""

    class Meta:
        verbose_name_plural = "Speciae"


class Subspecies(models.Model):
    """taxonomic Subspecies"""

    model_curi = "lara:organisms/Subspecies"
    subspecies_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False
    )
    name_display = models.TextField(
        help_text="human readable display name of the subspecies, can contain spaces"
    )
    name = models.TextField(blank=True, null=True, help_text="subspecies name, like ")
    name_common = models.TextField(
        blank=True, null=True, help_text="common subspecies name, like "
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_subspecies_traits_related",
        related_query_name="%(app_label)s_%(class)s_subspecies_traits",
        blank=True,
        help_text="General subspecies's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="General characteristics of this subspecies, use 'traits' for specific traits.",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    description = models.TextField(
        null=True, blank=True, help_text="description of this subspecies"
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for IRI
        """
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()
        if self.iri is None or self.iri == "":
            self.iri = f"{settings.LARA_PREFIXES['lara']}organisms/subspecies/{self.name.replace(' ', '_').lower().strip()}"
        self.datetime_last_modified = timezone.now()
        super().save(*args, **kwargs)

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return self.name or ""

    class Meta:
        verbose_name_plural = "Subspecieae"


class Variant(models.Model):
    """taxonomic Variant"""

    model_curi = "lara:organisms/Variant"
    variant_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the variant, can contain spaces"
    )
    name = models.TextField(blank=True, null=True, help_text="variant name, like ")
    name_common = models.TextField(
        blank=True, null=True, help_text="common variant name, like "
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_variant_traits_related",
        related_query_name="%(app_label)s_%(class)s_variant_traits",
        blank=True,
        help_text="General variant's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="General characteristics of this variant, use 'traits' for specific traits.",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    description = models.TextField(
        null=True, blank=True, help_text="description of this variant"
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for IRI
        """
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()
        if self.iri is None or self.iri == "":
            self.iri = f"{settings.LARA_PREFIXES['lara']}organisms/variant/{self.name.replace(' ', '_').lower().strip()}"
        self.datetime_last_modified = timezone.now()
        super().save(*args, **kwargs)

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return self.name or ""


class Cultivar(models.Model):
    """taxonomic Cultivar / Plant variety originating from cultivation / cell line"""

    model_curi = "lara:organisms/Cultivar"
    cultivar_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the cultivar, can contain spaces"
    )
    name = models.TextField(blank=True, null=True, help_text="cultivar name, like ")
    name_common = models.TextField(
        blank=True, null=True, help_text="common cultivar name, like "
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_cultivar_traits_related",
        related_query_name="%(app_label)s_%(class)s_cultivar_traits",
        blank=True,
        help_text="General cultivar's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="General characteristics of this cultivar, use 'traits' for specific traits.",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    description = models.TextField(
        null=True, blank=True, help_text="description of this cultivar"
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for IRI
        """
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()
        if self.iri is None or self.iri == "":
            self.iri = f"{settings.LARA_PREFIXES['lara']}organisms/cultivar/{self.name.replace(' ', '_').lower().strip()}"
        self.datetime_last_modified = timezone.now()
        super().save(*args, **kwargs)

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return self.name or ""


class Organism(models.Model):
    """organism / species / strain  or cell line, details should be in the organism's ontology"""

    model_curi = "lara:organisms/Organism"
    organism_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the organism, can contain spaces"
    )
    name = models.TextField(
        blank=True,
        null=True,
        help_text="organism name for referencing in code, should not contain whitespaces",
    )
    name_common = models.TextField(
        blank=True, null=True, help_text="common organism name, like 'elephant' "
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    # PAC-ID:  https://github.com/ApiniLabs/PAC-ID
    pac_id = models.URLField(
        blank=True,
        null=True,
        unique=True,
        help_text="Publicly Addressable Content IDentifier is a globally unique identifier that operates independently of a central registry.",
    )
    # taxonomy / classification
    domain = models.ForeignKey(
        Domain,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_organism_domain_related",
        related_query_name="%(app_label)s_%(class)s_organism_domain",
        blank=True,
        null=True,
        help_text="organism's domain name",
    )
    regnum = models.ForeignKey(
        Regnum,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_organism_regnum_related",
        related_query_name="%(app_label)s_%(class)s_organism_regnum",
        blank=True,
        null=True,
        help_text="organism's regnum name",
    )
    phylum = models.ForeignKey(
        Phylum,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_organism_phylum_related",
        related_query_name="%(app_label)s_%(class)s_organism_phylum",
        blank=True,
        null=True,
        help_text="organism's phylum name",
    )
    classis = models.ForeignKey(
        Classis,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_organism_classis_related",
        related_query_name="%(app_label)s_%(class)s_organism_classis",
        blank=True,
        null=True,
        help_text="organism's classis (class) name",
    )
    ordo = models.ForeignKey(
        Ordo,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_organism_ordo_related",
        related_query_name="%(app_label)s_%(class)s_organism_ordo",
        blank=True,
        null=True,
        help_text="organism's ordo name",
    )
    familia = models.ForeignKey(
        Familia,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_organism_familia_related",
        related_query_name="%(app_label)s_%(class)s_organism_familia",
        blank=True,
        null=True,
        help_text="organism's familia (family) name",
    )
    genus = models.ForeignKey(
        Genus,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_organism_genus_related",
        related_query_name="%(app_label)s_%(class)s_organism_genus",
        blank=True,
        null=True,
        help_text="organism's genus name",
    )
    species = models.ForeignKey(
        Species,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_organism_species_related",
        related_query_name="%(app_label)s_%(class)s_organism_species",
        blank=True,
        null=True,
        help_text="organism's species name",
    )
    subspecies = models.ForeignKey(
        Subspecies,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_organism_subspecies_related",
        related_query_name="%(app_label)s_%(class)s_organism_subspecies",
        blank=True,
        null=True,
        help_text="organism's subspecies name",
    )
    variant = models.ForeignKey(
        Variant,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_organism_variant_related",
        related_query_name="%(app_label)s_%(class)s_organism_variant",
        blank=True,
        null=True,
        help_text="organism's varian name",
    )
    cultivar = models.ForeignKey(
        Cultivar,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_organism_cultivar_related",
        related_query_name="%(app_label)s_%(class)s_organism_cultivar",
        blank=True,
        null=True,
        help_text="organism's cultivar name",
    )
    habitats = models.ManyToManyField(
        Habitat,
        related_name="%(app_label)s_%(class)s_organism_habitat_related",
        related_query_name="%(app_label)s_%(class)s_organism_habitat",
        blank=True,
        help_text="general organism's habitats",
    )
    organisms = models.ManyToManyField(
        "self",
        related_query_name="%(app_label)s_%(class)s_organism_organisms",
        blank=True,
        help_text="organisms involved in hybrid or chimera",
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_organism_traits_related",
        related_query_name="%(app_label)s_%(class)s_organism_traits",
        blank=True,
        help_text="general organism's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="General characteristics of this variant, use 'traits' for specific traits.",
    )
    # consider a link to lara-django-sequences to store the genome as a sequence, efficiency needs to be tested
    genome = models.URLField(
        blank=True,
        null=True,
        max_length=512,
        help_text="URL to organism's (reference) genome",
    )

    hybrid = models.BooleanField(
        default=False,
        help_text="Hybrid - organism with mixed genetic material like mule, liger/tigon, plant hybrids.",
    )
    chimera = models.BooleanField(
        default=False,
        help_text="Chimera - organisms with mixed cells from different species, like apple/pear tree.",
    )

    description = models.TextField(
        null=True, blank=True, help_text="description of the organism"
    )

    data_extra = models.ManyToManyField(
        ExtraData,
        related_name="%(app_label)s_%(class)s_organism_extradata_related",
        related_query_name="%(app_label)s_%(class)s_organism_extradata",
        blank=True,
        help_text="organism related extra ",
    )
    tags = models.ManyToManyField(
        Tag,
        related_name="%(app_label)s_%(class)s_organism_tags_related",
        related_query_name="%(app_label)s_%(class)s_organism_tags",
        blank=True,
        help_text="organism related tags",
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when organism data was last modified",
    )

    def save(self, *args, **kwargs):
        """
        Here we generate some default values for IRI
        """

        if self.name_display is None or self.name_display == "":
            try:
                self.name_display = (
                    f"{self.genus.name.capitalize()} {self.species.name}"
                )
            except Exception as e:
                logger.error(f"Error in generating name_display for organism: {e}")

        if self.name is None or self.name == "":
            if self.name_display is not None and self.name_display != "":
                self.name = self.name_display.replace(" ", "_").lower().strip()
            else:
                try:
                    self.name = f"{self.genus.name}_{self.species.name}".lower().strip()
                except Exception as e:
                    logger.error(f"Error in generating name for organism: {e}")

        if self.iri is None or self.iri == "":
            self.iri = f"{settings.LARA_PREFIXES['lara']}organisms/organism/{self.name.replace(' ', '_').lower().strip()}"

        self.datetime_last_modified = timezone.now()

        super().save(*args, **kwargs)

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return self.name or ""

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=[
                    "name",
                    "genus",
                    "species",
                    "subspecies",
                    "variant",
                    "cultivar",
                ],
                name="unique_organism",
            )
        ]


class OrganismsSubset(ItemSubsetAbstr):
    """User or group defined subset of organisms"""

    model_curi = "lara:organisms/OrganismsSubset"
    organismssubset_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False
    )
    organisms = models.ManyToManyField(
        Organism,
        related_name="%(app_label)s_%(class)s_organisms_related",
        related_query_name="%(app_label)s_%(class)s_organisms",
        blank=True,
        help_text="organisms in this subset",
        through="OrderedOrganismsSubset",
    )


class OrderedOrganismsSubset(models.Model):
    """Through model for the many-to-many relationship between OrganismsSubset and Organism.
    This class can be used to store the order of organisms in a subset"""

    orderedorganismssubset_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False
    )

    organism = models.ForeignKey(
        Organism,
        related_name="%(app_label)s_%(class)s_organism_related",
        related_query_name="%(app_label)s_%(class)s_organism",
        on_delete=models.CASCADE,
        help_text="organism in the subset",
    )

    organisms_subset = models.ForeignKey(
        OrganismsSubset,
        related_name="%(app_label)s_%(class)s_organismssubset_related",
        related_query_name="%(app_label)s_%(class)s_organismssubset",
        on_delete=models.CASCADE,
        help_text="subset of organisms",
    )

    order = models.IntegerField(help_text="order of organism in subset")

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["organism", "organisms_subset"],
                name="%(app_label)s_%(class)s_organisms_organisms_subset_unique",
            ),
        ]
        ordering = ("order",)



"""_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_myproj gRPC services*

:details: lara_django_myproj gRPC services.
         - 
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: - 
________________________________________________________________________
"""

## generated with django-socio-grpc generateprpcinterface lara_django_organisms  (LARA-version)

from django_socio_grpc import generics, mixins

from lara_django_base.grpc.mixins import AsyncUploadModelMixin, AsyncRetrieveByNameModelMixin, AsyncDownloadModelMixin

from .serializers import (
    ExtraDataProtoSerializer,
    HabitatClassProtoSerializer,
    HabitatParameterProtoSerializer,
    HabitatProtoSerializer,
    TraitProtoSerializer,
    DomainProtoSerializer,
    RegnumProtoSerializer,
    PhylumProtoSerializer,
    ClassisProtoSerializer,
    OrdoProtoSerializer,
    FamiliaProtoSerializer,
    GenusProtoSerializer,
    SpeciesProtoSerializer,
    SubspeciesProtoSerializer,
    VariantProtoSerializer,
    CultivarProtoSerializer,
    OrganismProtoSerializer,
    OrganismsSubsetProtoSerializer
)
from ..filters import (
    ExtraDataFilterSet,
    HabitatClassFilterSet,
    HabitatParameterFilterSet,
    HabitatFilterSet,
    TraitFilterSet,
    DomainFilterSet,
    RegnumFilterSet,
    PhylumFilterSet,
    ClassisFilterSet,
    OrdoFilterSet,
    FamiliaFilterSet,
    GenusFilterSet,
    SpeciesFilterSet,
    SubspeciesFilterSet,
    VariantFilterSet,
    CultivarFilterSet,
    OrganismFilterSet,
    OrganismsSubsetFilterSet
)

from lara_django_organisms.models import (
    ExtraData,
    HabitatClass,
    HabitatParameter,
    Habitat,
    Trait,
    Domain,
    Regnum,
    Phylum,
    Classis,
    Ordo,
    Familia,
    Genus,
    Species,
    Subspecies,
    Variant,
    Cultivar,
    Organism,
    OrganismsSubset
)

import lara_django_organisms_grpc.v1.lara_django_organisms_pb2 as lara_django_organisms_pb2

class ExtraDataService(generics.AsyncModelService, mixins.AsyncStreamModelMixin, AsyncUploadModelMixin,
    AsyncRetrieveByNameModelMixin,
    AsyncDownloadModelMixin,):
    pb2 = lara_django_organisms_pb2
    queryset = ExtraData.objects.all()
    serializer_class = ExtraDataProtoSerializer
    filterset_class = ExtraDataFilterSet


class HabitatClassService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    queryset = HabitatClass.objects.all()
    serializer_class = HabitatClassProtoSerializer
    filterset_class = HabitatClassFilterSet


class HabitatParameterService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    queryset = HabitatParameter.objects.all()
    serializer_class = HabitatParameterProtoSerializer
    filterset_class = HabitatParameterFilterSet


class HabitatService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    queryset = Habitat.objects.all()
    serializer_class = HabitatProtoSerializer
    filterset_class = HabitatFilterSet


class TraitService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    queryset = Trait.objects.all()
    serializer_class = TraitProtoSerializer
    filterset_class = TraitFilterSet


class DomainService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    queryset = Domain.objects.all()
    serializer_class = DomainProtoSerializer
    filterset_class = DomainFilterSet


class RegnumService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    queryset = Regnum.objects.all()
    serializer_class = RegnumProtoSerializer
    filterset_class = RegnumFilterSet


class PhylumService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    queryset = Phylum.objects.all()
    serializer_class = PhylumProtoSerializer
    filterset_class = PhylumFilterSet


class ClassisService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    queryset = Classis.objects.all()
    serializer_class = ClassisProtoSerializer
    filterset_class = ClassisFilterSet


class OrdoService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    queryset = Ordo.objects.all()
    serializer_class = OrdoProtoSerializer
    filterset_class = OrdoFilterSet


class FamiliaService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    queryset = Familia.objects.all()
    serializer_class = FamiliaProtoSerializer
    filterset_class = FamiliaFilterSet


class GenusService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    queryset = Genus.objects.all()
    serializer_class = GenusProtoSerializer
    filterset_class = GenusFilterSet


class SpeciesService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    queryset = Species.objects.all()
    serializer_class = SpeciesProtoSerializer
    filterset_class = SpeciesFilterSet


class SubspeciesService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    queryset = Subspecies.objects.all()
    serializer_class = SubspeciesProtoSerializer
    filterset_class = SubspeciesFilterSet


class VariantService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    queryset = Variant.objects.all()
    serializer_class = VariantProtoSerializer
    filterset_class = VariantFilterSet


class CultivarService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    queryset = Cultivar.objects.all()
    serializer_class = CultivarProtoSerializer
    filterset_class = CultivarFilterSet


class OrganismService(generics.AsyncModelService, mixins.AsyncStreamModelMixin, AsyncUploadModelMixin,
    AsyncRetrieveByNameModelMixin,
    AsyncDownloadModelMixin,):
    pb2 = lara_django_organisms_pb2
    queryset = Organism.objects.all()
    serializer_class = OrganismProtoSerializer
    filterset_class = OrganismFilterSet

class OrganismsSubsetService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    queryset = OrganismsSubset.objects.all()
    serializer_class = OrganismsSubsetProtoSerializer
    filterset_class = OrganismsSubsetFilterSet


"""_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_data filter app *

:details: lara_django_data filter app. 
         - 
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: - 
________________________________________________________________________
"""


from django_filters.rest_framework import FilterSet, CharFilter, DateRangeFilter

from .models import (
    ExtraData,
    HabitatClass,
    HabitatParameter,
    Habitat,
    Trait,
    Domain,
    Regnum,
    Phylum,
    Classis,
    Ordo,
    Familia,
    Genus,
    Species,
    Subspecies,
    Variant,
    Cultivar,
    Organism,
    OrganismsSubset
)


class ExtraDataFilterSet(FilterSet):
    class Meta:
        model = ExtraData
        fields = {
            "name_display": ["exact", "contains"],
            "name": ["exact", "contains"],
            "name_full": ["exact", "contains"],
            "iri": ["exact", "contains"],
            "url": ["exact", "contains"],
            "datetime_created": ["lt", "gt", "exact", "range"],
            "datetime_last_modified": ["lt", "gt", "exact", "range"],
            "description": ["exact", "contains"],
        }


class HabitatClassFilterSet(FilterSet):
    class Meta:
        model = HabitatClass
        fields = {
            "name": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class HabitatParameterFilterSet(FilterSet):
    class Meta:
        model = HabitatParameter
        fields = {
            "name": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class HabitatFilterSet(FilterSet):
    class Meta:
        model = Habitat
        fields = {
            "name": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class TraitFilterSet(FilterSet):
    class Meta:
        model = Trait
        fields = {
            "name": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class DomainFilterSet(FilterSet):
    class Meta:
        model = Domain
        fields = {
            "name": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class RegnumFilterSet(FilterSet):
    class Meta:
        model = Regnum
        fields = {
            "name": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class PhylumFilterSet(FilterSet):
    class Meta:
        model = Phylum
        fields = {
            "name": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class ClassisFilterSet(FilterSet):
    class Meta:
        model = Classis
        fields = {
            "name": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class OrdoFilterSet(FilterSet):
    class Meta:
        model = Ordo
        fields = {
            "name": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class FamiliaFilterSet(FilterSet):
    class Meta:
        model = Familia
        fields = {
            "name": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class GenusFilterSet(FilterSet):
    class Meta:
        model = Genus
        fields = {
            "name": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class SpeciesFilterSet(FilterSet):
    class Meta:
        model = Species
        fields = {
            "name": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class SubspeciesFilterSet(FilterSet):
    class Meta:
        model = Subspecies
        fields = {
            "name": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class VariantFilterSet(FilterSet):
    class Meta:
        model = Variant
        fields = {
            "name": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class CultivarFilterSet(FilterSet):
    class Meta:
        model = Cultivar
        fields = {
            "name": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class OrganismFilterSet(FilterSet):
    class Meta:
        model = Organism
        fields = {
            "name": ["exact", "contains"],
            "name_common": ["exact", "contains"],
            "domain__name": ["exact", "contains"],
            "familia__name": ["exact", "contains"],
            "genus__name": ["exact", "contains"],
            "species__name": ["exact", "contains"],
            "subspecies__name": ["exact", "contains"],
            "variant__name": ["exact", "contains"],
            "cultivar__name": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class OrganismsSubsetFilterSet(FilterSet):
    class Meta:
        model = OrganismsSubset
        fields = {
            "name": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


"""_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_myproj gRPC serializer*

:details: lara_django_myproj gRPC serializer.
         - 
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: - 
________________________________________________________________________
"""

## generated with django-socio-grpc generateprpcinterface lara_django_organisms  (LARA-version)

import logging

from django_socio_grpc import proto_serializers
from rest_framework.serializers import UUIDField, PrimaryKeyRelatedField
from lara_django_base.models import (
    Namespace,
    ItemStatus,
    DataType,
    MediaType,
    Namespace,
    Tag,
    PhysicalUnit,
    ExternalResource,
    Location,
    Currency,
)
from lara_django_people.models import (
    Entity,
    Group,
)

from lara_django_organisms.models import (
    ExtraData,
    HabitatClass,
    HabitatParameter,
    Habitat,
    Trait,
    Domain,
    Regnum,
    Phylum,
    Classis,
    Ordo,
    Familia,
    Genus,
    Species,
    Subspecies,
    Variant,
    Cultivar,
    Organism,
    OrganismsSubset,
)

import lara_django_organisms_grpc.v1.lara_django_organisms_pb2 as lara_django_organisms_pb2


class ExtraDataProtoSerializer(proto_serializers.ModelProtoSerializer):
    data_type = PrimaryKeyRelatedField(
        queryset=DataType.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    namespace = PrimaryKeyRelatedField(queryset=Namespace.objects.all(), pk_field=UUIDField(format="hex_verbose"))
    media_type = PrimaryKeyRelatedField(
        queryset=MediaType.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )

    class Meta:
        model = ExtraData
        proto_class = lara_django_organisms_pb2.ExtraDataResponse

        proto_class_list = lara_django_organisms_pb2.ExtraDataListResponse

        fields = "__all__"  # [data_type', namespace', URI', text', XML', JSON', bin', media_type', IRI', URL', description', file']


class HabitatClassProtoSerializer(proto_serializers.ModelProtoSerializer):
    class Meta:
        model = HabitatClass
        proto_class = lara_django_organisms_pb2.HabitatClassResponse

        proto_class_list = lara_django_organisms_pb2.HabitatClassListResponse

        fields = "__all__"  # [name', description']


class HabitatParameterProtoSerializer(proto_serializers.ModelProtoSerializer):
    unit = PrimaryKeyRelatedField(queryset=PhysicalUnit.objects.all(), pk_field=UUIDField(format="hex_verbose"))
    data_extra = PrimaryKeyRelatedField(
        queryset=ExtraData.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )

    class Meta:
        model = HabitatParameter
        proto_class = lara_django_organisms_pb2.HabitatParameterResponse

        proto_class_list = lara_django_organisms_pb2.HabitatParameterListResponse

        fields = "__all__"  # [name', unit', description']


class HabitatProtoSerializer(proto_serializers.ModelProtoSerializer):
    habitat_class = PrimaryKeyRelatedField(
        queryset=HabitatClass.objects.all(), pk_field=UUIDField(format="hex_verbose")
    )
    parameters = PrimaryKeyRelatedField(
        queryset=HabitatParameter.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )

    class Meta:
        model = Habitat
        proto_class = lara_django_organisms_pb2.HabitatResponse

        proto_class_list = lara_django_organisms_pb2.HabitatListResponse

        fields = "__all__"  # [name', description']


class TraitProtoSerializer(proto_serializers.ModelProtoSerializer):
    unit = PrimaryKeyRelatedField(
        queryset=PhysicalUnit.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )

    class Meta:
        model = Trait
        proto_class = lara_django_organisms_pb2.TraitResponse

        proto_class_list = lara_django_organisms_pb2.TraitListResponse

        fields = "__all__"  # [feature', unit', description']


class DomainProtoSerializer(proto_serializers.ModelProtoSerializer):
    traits = PrimaryKeyRelatedField(
        queryset=Trait.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    class Meta:
        model = Domain
        proto_class = lara_django_organisms_pb2.DomainResponse

        proto_class_list = lara_django_organisms_pb2.DomainListResponse

        fields = "__all__"  # [name', description']


class RegnumProtoSerializer(proto_serializers.ModelProtoSerializer):
    traits = PrimaryKeyRelatedField(
        queryset=Trait.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    class Meta:
        model = Regnum
        proto_class = lara_django_organisms_pb2.RegnumResponse

        proto_class_list = lara_django_organisms_pb2.RegnumListResponse

        fields = "__all__"  # [name', description']


class PhylumProtoSerializer(proto_serializers.ModelProtoSerializer):
    traits = PrimaryKeyRelatedField(
        queryset=Trait.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    class Meta:
        model = Phylum
        proto_class = lara_django_organisms_pb2.PhylumResponse

        proto_class_list = lara_django_organisms_pb2.PhylumListResponse

        fields = "__all__"  # [name', name_common', characteristics', description']


class ClassisProtoSerializer(proto_serializers.ModelProtoSerializer):
    traits = PrimaryKeyRelatedField(
        queryset=Trait.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    class Meta:
        model = Classis
        proto_class = lara_django_organisms_pb2.ClassisResponse

        proto_class_list = lara_django_organisms_pb2.ClassisListResponse

        fields = "__all__"  # [name', name_common', characteristics', description']


class OrdoProtoSerializer(proto_serializers.ModelProtoSerializer):
    traits = PrimaryKeyRelatedField(
        queryset=Trait.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    class Meta:
        model = Ordo
        proto_class = lara_django_organisms_pb2.OrdoResponse

        proto_class_list = lara_django_organisms_pb2.OrdoListResponse

        fields = "__all__"  # [name', name_common', characteristics', description']


class FamiliaProtoSerializer(proto_serializers.ModelProtoSerializer):
    traits = PrimaryKeyRelatedField(
        queryset=Trait.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    class Meta:
        model = Familia
        proto_class = lara_django_organisms_pb2.FamiliaResponse

        proto_class_list = lara_django_organisms_pb2.FamiliaListResponse

        fields = "__all__"  # [name', name_common', characteristics', description']


class GenusProtoSerializer(proto_serializers.ModelProtoSerializer):
    traits = PrimaryKeyRelatedField(
        queryset=Trait.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    class Meta:
        model = Genus
        proto_class = lara_django_organisms_pb2.GenusResponse

        proto_class_list = lara_django_organisms_pb2.GenusListResponse

        fields = "__all__"  # [name', name_common', characteristics', description']


class SpeciesProtoSerializer(proto_serializers.ModelProtoSerializer):
    traits = PrimaryKeyRelatedField(
        queryset=Trait.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    class Meta:
        model = Species
        proto_class = lara_django_organisms_pb2.SpeciesResponse

        proto_class_list = lara_django_organisms_pb2.SpeciesListResponse

        fields = "__all__"  # [name', name_common', characteristics', description']


class SubspeciesProtoSerializer(proto_serializers.ModelProtoSerializer):
    traits = PrimaryKeyRelatedField(
        queryset=Trait.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    class Meta:
        model = Subspecies
        proto_class = lara_django_organisms_pb2.SubspeciesResponse

        proto_class_list = lara_django_organisms_pb2.SubspeciesListResponse

        fields = "__all__"  # [name', name_common', characteristics', description']


class VariantProtoSerializer(proto_serializers.ModelProtoSerializer):
    traits = PrimaryKeyRelatedField(
        queryset=Trait.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    class Meta:
        model = Variant
        proto_class = lara_django_organisms_pb2.VariantResponse

        proto_class_list = lara_django_organisms_pb2.VariantListResponse

        fields = "__all__"  # [name', name_common', characteristics', description']


class CultivarProtoSerializer(proto_serializers.ModelProtoSerializer):
    traits = PrimaryKeyRelatedField(
        queryset=Trait.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    class Meta:
        model = Cultivar
        proto_class = lara_django_organisms_pb2.CultivarResponse

        proto_class_list = lara_django_organisms_pb2.CultivarListResponse

        fields = "__all__"  # [name', name_common', characteristics', description']


class OrganismProtoSerializer(proto_serializers.ModelProtoSerializer):
    domain = PrimaryKeyRelatedField(
        queryset=Domain.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    regnum = PrimaryKeyRelatedField(
        queryset=Regnum.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    phylum = PrimaryKeyRelatedField(
        queryset=Phylum.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    classis = PrimaryKeyRelatedField(
        queryset=Classis.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    ordo = PrimaryKeyRelatedField(
        queryset=Ordo.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    familia = PrimaryKeyRelatedField(
        queryset=Familia.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    genus = PrimaryKeyRelatedField(queryset=Genus.objects.all(), pk_field=UUIDField(format="hex_verbose"))
    species = PrimaryKeyRelatedField(queryset=Species.objects.all(), pk_field=UUIDField(format="hex_verbose"))
    subspecies = PrimaryKeyRelatedField(
        queryset=Subspecies.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    variant = PrimaryKeyRelatedField(
        queryset=Variant.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    cultivar = PrimaryKeyRelatedField(
        queryset=Cultivar.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    habitats = PrimaryKeyRelatedField(
        queryset=Habitat.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    traits = PrimaryKeyRelatedField(
        queryset=Trait.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    organisms = PrimaryKeyRelatedField(
        queryset=Organism.objects.all(), pk_field=UUIDField(format="hex_verbose"), many=True, required=False, allow_null=True
    )
    data_extra = PrimaryKeyRelatedField(
        queryset=ExtraData.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    tags = PrimaryKeyRelatedField(
        queryset=Tag.objects.all(), pk_field=UUIDField(format="hex_verbose"), many=True, required=False, allow_null=True
    )

    class Meta:
        model = Organism
        proto_class = lara_django_organisms_pb2.OrganismResponse

        proto_class_list = lara_django_organisms_pb2.OrganismListResponse

        fields = "__all__"  # [name', name_common', domain', regnum', phylum', classis', ordo', familia', genus', species', subspecies', variant', cultivar', habitat', description']

class OrganismsSubsetProtoSerializer(proto_serializers.ModelProtoSerializer):
    namespace = PrimaryKeyRelatedField(queryset=Namespace.objects.all(), pk_field=UUIDField(format="hex_verbose"))
    entity = PrimaryKeyRelatedField(queryset=Entity.objects.all(), pk_field=UUIDField(format="hex_verbose"))
    group = PrimaryKeyRelatedField(
        queryset=Group.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    organisms = PrimaryKeyRelatedField(
        queryset=Organism.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
   
    tags = PrimaryKeyRelatedField(
        queryset=Tag.objects.all(), pk_field=UUIDField(format="hex_verbose"), many=True, required=False, allow_null=True
    )

    class Meta:
        model = OrganismsSubset
        proto_class = lara_django_organisms_pb2.OrganismsSubsetResponse

        proto_class_list = lara_django_organisms_pb2.OrganismsSubsetListResponse

        fields = "__all__"  # [name', description']


"""_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_organisms app *

:details: lara_django_organisms app configuration. 
         This provides a generic django app configuration mechanism.
         For more details see:
         https://docs.djangoproject.com/en/4.0/ref/applications/
         - 
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: - 
________________________________________________________________________
"""


from django.apps import AppConfig


class LaraDjangoOrganismsConfig(AppConfig):
    name = 'lara_django_organisms'
    url_path = 'organisms/'
    # enter a verbose name for your app: lara_django_organisms here - this will be used in the admin interface
    verbose_name = 'LARA-django Organisms'
    # lara_app_icon = 'lara_django_organisms_icon.svg'  # this will be used to display an icon, e.g. in the main LARA menu.
    lara_app_color = 'lime'  # this will be used to highlight the app in the main LARA sidebar and app page

    def ready(self):
        # add urlpatterns 
        from django.urls import path, include
        from lara_django.urls import urlpatterns

        urlpatterns += [ 
            path(self.url_path, include('lara_django_organisms.urls', namespace='lara_django_organisms')),
        ]



"""_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_organisms admin *

:details: lara_django_organisms admin module admin backend configuration.
         - 
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: - run "lara-django-dev forms_generator -c lara_django_organisms > forms.py" to update this file
________________________________________________________________________
"""
# django crispy forms s. https://github.com/django-crispy-forms/django-crispy-forms for []
# generated with django-extensions forms_generator -c  [] > forms.py (LARA-version)

from django import forms
from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout, Submit, Row, Column

from .models import (
    ExtraData,
    HabitatClass,
    HabitatParameter,
    Habitat,
    Trait,
    Domain,
    Regnum,
    Phylum,
    Classis,
    Ordo,
    Familia,
    Genus,
    Species,
    Subspecies,
    Variant,
    Cultivar,
    Organism,
)


class ExtraDataCreateForm(forms.ModelForm):
    class Meta:
        model = ExtraData
        fields = (
            "data_type",
            "name",
            "name_full",
            "name_display",
            "namespace",
            "version",
          
            "hash_sha256",
            "data_json",
            "media_type",
            "iri",
            "url",
            "description",
            "file",
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "data_type",
            "name",
            "name_full",
            "name_display",
            "namespace",
            "version",
          
            "hash_sha256",
            "data_json",
            "media_type",
            "iri",
            "url",
            "description",
            "file",
            Submit("submit", "Create"),
        )


class ExtraDataUpdateForm(forms.ModelForm):
    class Meta:
        model = ExtraData
        fields = (
            "data_type",
            "name",
            "name_full",
            "name_display",
            "namespace",
            "version",
          
            "hash_sha256",
            "data_json",
            "media_type",
            "iri",
            "url",
            "description",
            "file",
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "data_type",
            "name",
            "name_full",
            "name_display",
            "namespace",
            "version",
          
            "hash_sha256",
            "data_json",
            "media_type",
            "iri",
            "url",
            "description",
            "file",
            Submit("submit", "Update"),
        )


class HabitatClassCreateForm(forms.ModelForm):
    class Meta:
        model = HabitatClass
        fields = ("name", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("name", "iri", "description", Submit("submit", "Create"))


class HabitatClassUpdateForm(forms.ModelForm):
    class Meta:
        model = HabitatClass
        fields = ("name", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("name", "iri", "description", Submit("submit", "Update"))


class HabitatParameterCreateForm(forms.ModelForm):
    class Meta:
        model = HabitatParameter
        fields = ("name", "unit", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("name", "unit", "iri", "description", Submit("submit", "Create"))


class HabitatParameterUpdateForm(forms.ModelForm):
    class Meta:
        model = HabitatParameter
        fields = ("name", "unit", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("name", "unit", "iri", "description", Submit("submit", "Update"))


class HabitatCreateForm(forms.ModelForm):
    class Meta:
        model = Habitat
        fields = ("name", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("name", "description", Submit("submit", "Create"))


class HabitatUpdateForm(forms.ModelForm):
    class Meta:
        model = Habitat
        fields = ("name", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("name", "description", Submit("submit", "Update"))


class TraitCreateForm(forms.ModelForm):
    class Meta:
        model = Trait
        fields = ("name", "unit", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("name", "unit", "iri", "description", Submit("submit", "Create"))


class TraitUpdateForm(forms.ModelForm):
    class Meta:
        model = Trait
        fields = ("name", "unit", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("name", "unit", "iri", "description", Submit("submit", "Update"))


class DomainCreateForm(forms.ModelForm):
    class Meta:
        model = Domain
        fields = ("name", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("name", "iri", "description", Submit("submit", "Create"))


class DomainUpdateForm(forms.ModelForm):
    class Meta:
        model = Domain
        fields = ("name", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("name", "iri", "description", Submit("submit", "Update"))


class RegnumCreateForm(forms.ModelForm):
    class Meta:
        model = Regnum
        fields = ("name", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("name", "iri", "description", Submit("submit", "Create"))


class RegnumUpdateForm(forms.ModelForm):
    class Meta:
        model = Regnum
        fields = ("name", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout("name", "iri", "description", Submit("submit", "Update"))


class PhylumCreateForm(forms.ModelForm):
    class Meta:
        model = Phylum
        fields = ("name", "name_common", "characteristics", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "name_common", "characteristics", "iri", "description", Submit("submit", "Create")
        )


class PhylumUpdateForm(forms.ModelForm):
    class Meta:
        model = Phylum
        fields = ("name", "name_common", "characteristics", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "name_common", "characteristics", "iri", "description", Submit("submit", "Update")
        )


class ClassisCreateForm(forms.ModelForm):
    class Meta:
        model = Classis
        fields = ("name", "name_common", "characteristics", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "name_common", "characteristics", "iri", "description", Submit("submit", "Create")
        )


class ClassisUpdateForm(forms.ModelForm):
    class Meta:
        model = Classis
        fields = ("name", "name_common", "characteristics", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "name_common", "characteristics", "iri", "description", Submit("submit", "Update")
        )


class OrdoCreateForm(forms.ModelForm):
    class Meta:
        model = Ordo
        fields = ("name", "name_common", "characteristics", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "name_common", "characteristics", "iri", "description", Submit("submit", "Create")
        )


class OrdoUpdateForm(forms.ModelForm):
    class Meta:
        model = Ordo
        fields = ("name", "name_common", "characteristics", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "name_common", "characteristics", "iri", "description", Submit("submit", "Update")
        )


class FamiliaCreateForm(forms.ModelForm):
    class Meta:
        model = Familia
        fields = ("name", "name_common", "characteristics", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "name_common", "characteristics", "iri", "description", Submit("submit", "Create")
        )


class FamiliaUpdateForm(forms.ModelForm):
    class Meta:
        model = Familia
        fields = ("name", "name_common", "characteristics", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "name_common", "characteristics", "iri", "description", Submit("submit", "Update")
        )


class GenusCreateForm(forms.ModelForm):
    class Meta:
        model = Genus
        fields = ("name", "name_common", "characteristics", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "name_common", "characteristics", "iri", "description", Submit("submit", "Create")
        )


class GenusUpdateForm(forms.ModelForm):
    class Meta:
        model = Genus
        fields = ("name", "name_common", "characteristics", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "name_common", "characteristics", "iri", "description", Submit("submit", "Update")
        )


class SpeciesCreateForm(forms.ModelForm):
    class Meta:
        model = Species
        fields = ("name", "name_common", "characteristics", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "name_common", "characteristics", "iri", "description", Submit("submit", "Create")
        )


class SpeciesUpdateForm(forms.ModelForm):
    class Meta:
        model = Species
        fields = ("name", "name_common", "characteristics", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "name_common", "characteristics", "iri", "description", Submit("submit", "Update")
        )


class SubspeciesCreateForm(forms.ModelForm):
    class Meta:
        model = Subspecies
        fields = ("name", "name_common", "characteristics", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "name_common", "characteristics", "iri", "description", Submit("submit", "Create")
        )


class SubspeciesUpdateForm(forms.ModelForm):
    class Meta:
        model = Subspecies
        fields = ("name", "name_common", "characteristics", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "name_common", "characteristics", "iri", "description", Submit("submit", "Update")
        )


class VariantCreateForm(forms.ModelForm):
    class Meta:
        model = Variant
        fields = ("name", "name_common", "characteristics", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "name_common", "characteristics", "iri", "description", Submit("submit", "Create")
        )


class VariantUpdateForm(forms.ModelForm):
    class Meta:
        model = Variant
        fields = ("name", "name_common", "characteristics", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "name_common", "characteristics", "iri", "description", Submit("submit", "Update")
        )


class CultivarCreateForm(forms.ModelForm):
    class Meta:
        model = Cultivar
        fields = ("name", "name_common", "characteristics", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "name_common", "characteristics", "iri", "description", Submit("submit", "Create")
        )


class CultivarUpdateForm(forms.ModelForm):
    class Meta:
        model = Cultivar
        fields = ("name", "name_common", "characteristics", "iri", "description")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name", "name_common", "characteristics", "iri", "description", Submit("submit", "Update")
        )


class OrganismCreateForm(forms.ModelForm):
    class Meta:
        model = Organism
        fields = (
            "name",
            "name_common",
            "domain",
            "regnum",
            "phylum",
            "classis",
            "ordo",
            "familia",
            "genus",
            "species",
            "subspecies",
            "variant",
            "cultivar",
            "habitats",
            "traits",
            "iri",
            "description",
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name",
            "name_common",
            "domain",
            "regnum",
            "phylum",
            "classis",
            "ordo",
            "familia",
            "genus",
            "species",
            "subspecies",
            "variant",
            "cultivar",
            "habitats",
            "traits",
            "iri",
            "description",
            Submit("submit", "Create"),
        )


class OrganismUpdateForm(forms.ModelForm):
    class Meta:
        model = Organism
        fields = (
            "name",
            "name_common",
            "domain",
            "regnum",
            "phylum",
            "classis",
            "ordo",
            "familia",
            "genus",
            "species",
            "subspecies",
            "variant",
            "cultivar",
            "habitats",
            "traits",
            "iri",
            "description",
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            "name",
            "name_common",
            "domain",
            "regnum",
            "phylum",
            "classis",
            "ordo",
            "familia",
            "genus",
            "species",
            "subspecies",
            "variant",
            "cultivar",
            "habitats",
            "traits",
            "iri",
            "description",
            Submit("submit", "Update"),
        )


# from .forms import ExtraDataCreateForm, HabitatClassCreateForm, HabitatParameterClassCreateForm, HabitatCreateForm, FeatureClassCreateForm, DomainCreateForm, RegnumCreateForm, PhylumCreateForm, ClassisCreateForm, OrdoCreateForm, FamiliaCreateForm, GenusCreateForm, SpeciesCreateForm, SubspeciesCreateForm, VariantCreateForm, CultivarCreateForm, OrganismCreateFormExtraDataUpdateForm, HabitatClassUpdateForm, HabitatParameterClassUpdateForm, HabitatUpdateForm, FeatureClassUpdateForm, DomainUpdateForm, RegnumUpdateForm, PhylumUpdateForm, ClassisUpdateForm, OrdoUpdateForm, FamiliaUpdateForm, GenusUpdateForm, SpeciesUpdateForm, SubspeciesUpdateForm, VariantUpdateForm, CultivarUpdateForm, OrganismUpdateForm



"""
_____________________________________________________________________.

:PROJECT: LARAsuite

*lara_django_organisms Model Context Protocol (MCP) handler *

:details: lara_django_organisms Model Context Protocol (MCP) handler.

:authors: mark doerr <mark.doerr@uni-greifswald.de>

________________________________________________________________________
"""

# from mcp_server import ModelQueryToolset

# from .models import Bird


# class BirdQueryTool(ModelQueryToolset):
#     model = Bird

#     def get_queryset(self):
#         """
#         Returns a queryset of all birds with a location.

#         This is used to filter the queryset for the admin interface.
#         self.request can be used to filter the queryset
#         """
#         return super().get_queryset().filter(location__isnull=False)



"""_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_organisms views *

:details: lara_django_organisms views module.
         - add app specific urls here
         - 
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: -
________________________________________________________________________
"""

from dataclasses import dataclass, field
from typing import List

from django.shortcuts import get_object_or_404, render, redirect
from django.http import HttpResponseRedirect, HttpResponse
from django.urls import reverse

from django.views.generic import DetailView, ListView, CreateView, UpdateView, DeleteView

from django_tables2 import SingleTableView

from .forms import OrganismCreateForm, OrganismUpdateForm
from .tables import OrganismTable
from .models import Organism

# Create your  lara_django_organisms views here.


@dataclass
class OrganismsMenu:
    menu_items:  List[dict] = field(default_factory=lambda: [
        {'name': 'Organisms',
         'path': 'lara_django_organisms:organism-list'},
        {'name': 'Organisms-Store',
         'path': 'lara_django_organisms_store:organism-list'},
    ])



class OrganismsSingleTableView(SingleTableView):
    model = Organism
    table_class = OrganismTable

    #fields = ('name', 'name_full', 'URL', 'pid', 'iri', 'manufacturer', 'model_no', 'product_type', 'type_barcode', 'product_no', 'weight', 'specifications', 'spec_json', 'spec_doc', 'brochure', 'quickstart', 'manual', 'service_manual', 'icon', 'image', 'description', 'organism_id', 'organism_class', 'shape')

    template_name = 'lara_django_organisms/list.html'
    success_url = '/organisms/organism/list'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['section_title'] = "Organisms - List"
        context['create_link'] = 'lara_django_organisms:organism-create'
        context['menu_items'] = OrganismsMenu().menu_items
        return context


class OrganismsDetailView(DetailView):
    model = Organism

    template_name = 'lara_django_organisms/organism_detail.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['section_title'] = "Organism - Details"
        context['update_link'] = 'lara_django_organisms:organism-update'
        context['menu_items'] = OrganismsMenu().menu_items
        return context


class OrganismsCreateView(CreateView):
    model = Organism

    template_name = 'lara_django_organisms/create_form.html'
    form_class = OrganismCreateForm
    success_url = '/organisms/organism/list'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['section_title'] = "Organism - Create"
        return context


class OrganismsUpdateView(UpdateView):
    model = Organism

    template_name = 'lara_django_organisms/update_form.html'
    form_class = OrganismUpdateForm
    success_url = '/organisms/organism/list'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['section_title'] = "Organism - Update"
        context['delete_link'] = 'lara_django_organisms:organism-delete'
        return context


class OrganismsDeleteView(DeleteView):
    model = Organism

    template_name = 'lara_django_organisms/delete_form.html'
    success_url = '/organisms/organism/list'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['section_title'] = "Organism - Delete"
        context['delete_link'] = 'lara_django_organisms:organism-delete'
        return context



"""_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_organisms urls *

:details: lara_django_organisms urls module.
         - add app specific urls here
         - 
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: -
________________________________________________________________________
"""

from . import views
from django.conf import settings
from django.conf.urls.static import static
from django.urls import path, include
from django.views.generic import TemplateView
from django_socio_grpc.settings import grpc_settings

from . import views

# Add your {cookiecutter.project_slug}} urls here.


# !! this sets the apps namespace to be used in the template
app_name = "lara_django_organisms"

# companies and institutions should also be added
# the 'name' attribute is used in templates to address the url independent of the view

urlpatterns = [
    path('organism/list/', views.OrganismsSingleTableView.as_view(), name='organism-list'),
    path('organism/create/', views.OrganismsCreateView.as_view(), name='organism-create'),
    path('organism/update/<uuid:pk>', views.OrganismsUpdateView.as_view(), name='organism-update'),
    path('organism/delete/<uuid:pk>', views.OrganismsDeleteView.as_view(), name='organism-delete'),
    path('organism/<uuid:pk>/', views.OrganismsDetailView.as_view(), name='organism-detail'),
    path('', views.OrganismsSingleTableView.as_view(), name='organisms-root'),
]  

# register handlers in settings
grpc_settings.user_settings["GRPC_HANDLERS"] += [
    "lara_django_organisms.grpc.handlers.grpc_handlers"]
